@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_sync.c,v 1.6 2002/07/11 14:26:41 wensong Exp $
+ * Version:     $Id: ip_vs_sync.c,v 1.7 2002/08/10 04:32:35 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *
@@ -132,7 +132,7 @@ static inline struct ip_vs_sync_buff * sb_dequeue(void)
 {
 	struct ip_vs_sync_buff *sb;
 
-	spin_lock(&ip_vs_sync_lock);
+	spin_lock_bh(&ip_vs_sync_lock);
 	if (list_empty(&ip_vs_sync_queue)) {
 		sb = NULL;
 	} else {
@@ -141,7 +141,7 @@ static inline struct ip_vs_sync_buff * sb_dequeue(void)
 				list);
 		list_del(&sb->list);
 	}
-	spin_unlock(&ip_vs_sync_lock);
+	spin_unlock_bh(&ip_vs_sync_lock);
 
 	return sb;
 }
@@ -173,20 +173,21 @@ static inline void ip_vs_sync_buff_release(struct ip_vs_sync_buff *sb)
 
 /*
  *	Get the current sync buffer if it has been created for more
- *	than the specified time.
+ *	than the specified time or the specified time is zero.
  */
 static inline struct ip_vs_sync_buff *
 get_curr_sync_buff(unsigned long time)
 {
 	struct ip_vs_sync_buff *sb;
 
-	spin_lock(&curr_sb_lock);
-	if (curr_sb && (jiffies - curr_sb->firstuse > time)) {
+	spin_lock_bh(&curr_sb_lock);
+	if (curr_sb &&
+	    (jiffies - curr_sb->firstuse > time || time == 0)) {
 		sb = curr_sb;
 		curr_sb = NULL;
 	} else
 		sb = NULL;
-	spin_unlock(&curr_sb_lock);
+	spin_unlock_bh(&curr_sb_lock);
 	return sb;
 }
 
@@ -210,7 +211,7 @@ void ip_vs_sync_conn(struct ip_vs_conn *cp)
 		}
 	}
 
-	len = cp->flags&IP_VS_CONN_F_SEQ_MASK ? FULL_CONN_SIZE :
+	len = (cp->flags & IP_VS_CONN_F_SEQ_MASK) ? FULL_CONN_SIZE :
 		SIMPLE_CONN_SIZE;
 	m = curr_sb->mesg;
 	s = (struct ip_vs_sync_conn *)curr_sb->head;
@@ -561,7 +562,7 @@ ip_vs_receive(struct socket *sock, char *buffer, const size_t buflen)
 static int errno;
 
 static DECLARE_WAIT_QUEUE_HEAD(sync_wait);
-static int sync_pid = 0;
+static pid_t sync_pid = 0;
 
 static DECLARE_WAIT_QUEUE_HEAD(stop_sync_wait);
 static int stop_sync = 0;
@@ -586,7 +587,7 @@ static void sync_master_loop(void)
 			ip_vs_sync_buff_release(sb);
 		}
 
-		/* check if entries stay in curr_sb for a while */
+		/* check if entries stay in curr_sb for 2 seconds */
 		if ((sb = get_curr_sync_buff(2*HZ))) {
 			m = sb->mesg;
 			if (ip_vs_send_async(sock, (char *)m,
@@ -603,16 +604,15 @@ static void sync_master_loop(void)
 		__set_current_state(TASK_RUNNING);
 	}
 
-	/* cleanup the sync_buff queue and the current sync_buff */
+	/* clean up the sync_buff queue */
 	while ((sb=sb_dequeue())) {
 		ip_vs_sync_buff_release(sb);
 	}
-	spin_lock(&curr_sb_lock);
-	if (curr_sb) {
-		ip_vs_sync_buff_release(curr_sb);
-		curr_sb = NULL;
+
+	/* clean up the current sync_buff */
+	if ((sb = get_curr_sync_buff(0))) {
+		ip_vs_sync_buff_release(sb);
 	}
-	spin_unlock(&curr_sb_lock);
 
 	/* release the sending multicast socket */
 	sock_release(sock);
@@ -643,7 +643,11 @@ static void sync_backup_loop(void)
 				IP_VS_ERR("receiving message error\n");
 				break;
 			}
+			/* disable bottom half, because it accessed the data
+			   shared by softirq while getting/creating conns */
+			local_bh_disable();
 			ip_vs_process_message(buf, len);
+			local_bh_enable();
 		}
 
 		if (stop_sync)
@@ -662,20 +666,17 @@ static void sync_backup_loop(void)
 }
 
 
-static int sync_thread(void *data)
+static int sync_thread(void *startup)
 {
 	DECLARE_WAITQUEUE(wait, current);
 	mm_segment_t oldmm;
+	int state;
 
-	MOD_INC_USE_COUNT;
 	daemonize();
 
 	oldmm = get_fs();
 	set_fs(KERNEL_DS);
 
-	sync_pid = current->pid;
-	IP_VS_INFO("sync thread started.\n");
-
 	if (ip_vs_sync_state == IP_VS_STATE_MASTER)
 		sprintf(current->comm, "ipvs syncmaster");
 	else if (ip_vs_sync_state == IP_VS_STATE_BACKUP)
@@ -693,41 +694,47 @@ static int sync_thread(void *data)
 	mcast_addr.sin_addr.s_addr = htonl(IP_VS_SYNC_GROUP);
 
 	add_wait_queue(&sync_wait, &wait);
+	state = ip_vs_sync_state;
+	complete((struct completion *)startup);
 
 	/* processing master/backup loop here */
-	if (ip_vs_sync_state == IP_VS_STATE_MASTER)
+	if (state == IP_VS_STATE_MASTER)
 		sync_master_loop();
-	else if (ip_vs_sync_state == IP_VS_STATE_BACKUP)
+	else if (state == IP_VS_STATE_BACKUP)
 		sync_backup_loop();
 	else IP_VS_BUG();
 
 	remove_wait_queue(&sync_wait, &wait);
 
 	/* thread exits */
-	sync_pid = 0;
-	IP_VS_INFO("sync thread stopped!\n");
-
 	stop_sync = 0;
 	wake_up(&stop_sync_wait);
 
 	set_fs(oldmm);
-	MOD_DEC_USE_COUNT;
 
 	return 0;
 }
 
 int start_sync_thread(int state, char *mcast_ifn)
 {
+	DECLARE_COMPLETION(startup);
+
 	if (sync_pid)
 		return -EEXIST;
 
+	MOD_INC_USE_COUNT;
+	IP_VS_DBG(7, "%s: pid %d\n", __FUNCTION__, current->pid);
 	IP_VS_DBG(7, "Each ip_vs_sync_conn entry need %d bytes\n",
 		  sizeof(struct ip_vs_sync_conn));
+
 	ip_vs_sync_state = state;
 	strcpy(ip_vs_mcast_ifn, mcast_ifn);
 
-	if (kernel_thread(sync_thread, NULL, 0) < 0)
+	if ((sync_pid = kernel_thread(sync_thread, &startup, 0)) < 0)
 		IP_VS_BUG();
+	wait_for_completion(&startup);
+
+	IP_VS_INFO("sync thread started.\n");
 	return 0;
 }
 
@@ -735,10 +742,12 @@ int start_sync_thread(int state, char *mcast_ifn)
 int stop_sync_thread(void)
 {
 	DECLARE_WAITQUEUE(wait, current);
+	pid_t pid;
 
 	if (!sync_pid)
 		return -ESRCH;
 
+	IP_VS_DBG(7, "%s: pid %d\n", __FUNCTION__, current->pid);
 	IP_VS_INFO("stopping sync thread %d ...\n", sync_pid);
 
 	__set_current_state(TASK_UNINTERRUPTIBLE);
@@ -750,13 +759,19 @@ int stop_sync_thread(void)
 	__set_current_state(TASK_RUNNING);
 	remove_wait_queue(&stop_sync_wait, &wait);
 
-	/* it is probably not correct here. we need use sys_wait4, but need
-	to patch the kernel to export sys_wait4, check it later...
-	sys_wait4(sync_pid, NULL, __WCLONE, NULL); */
-	waitpid(sync_pid, NULL, __WCLONE);
+	/* When the same process start/stop the sync thread (in this case
+	   the parent process of the sync thread is this process, instead
+	   of the init process), must reap the sync thread here. */
+	pid = waitpid(sync_pid, NULL, __WCLONE|WNOHANG);
+	if (pid > 0 && pid != sync_pid)
+		IP_VS_ERR("waitpid bug!\n");
 
 	if (stop_sync)
 		IP_VS_BUG();
 
+	IP_VS_INFO("sync thread stopped!\n");
+	sync_pid = 0;
+	MOD_DEC_USE_COUNT;
+
 	return 0;
 }
